import itertools
import os

import csaps
import matplotlib.pyplot as plt
import numpy as np

PLOT_STYLES = list(itertools.product(
    ('solid', 'dashed', 'dashdot', 'dotted',),
    ('r', 'g', 'b', 'c', 'm', 'y', 'k',)
))


def plot_betting_distribution(model_name, betting_data):
    fig = plt.figure(figsize=(9, 4))
    subplotIdx = 1
    for season, data in betting_data.items():
        ax = plt.subplot(1, 2, subplotIdx)
        ax.set_ylabel("Odds")
        ax.set_xlabel("Predicted probability")
        ax.set_xlim(0, 1)
        betting_history = data["history"]
        wins = list(filter(lambda b: b['result'] == b['bet'], betting_history))
        losses = list(filter(lambda b: b['result'] != b['bet'], betting_history))

        ax.scatter([b['probability'] for b in losses], [b['odds'] for b in losses], c='r', marker='o', s=8, label="Losing bet")
        ax.scatter([b['probability'] for b in wins], [b['odds'] for b in wins], c='g', marker='^', s=10, label="Winning bet")
        ax.title.set_text(f"{season}/{int(season)+1}")
        ax.legend()
        ax.grid(True)
        subplotIdx += 1
    filename = f"{model_name.replace(' ', '')}-distribution"
    print(filename)
    plt.tight_layout()
    plt.savefig(f"plots/{filename}.png")#, bbox_inches="tight")
    plt.show()


def plot_prediction_success(model_name, betting_data):
    fig = plt.figure(figsize=(9, 4))
    subplotIdx = 1
    for season, data in betting_data.items():
        ax = plt.subplot(1, 2, subplotIdx)
        ax.set_xlabel("Predicted probability")
        ax.set_ylabel("Outcome distribution")
        ax.set_xlim(0.08, 1)
        ax.set_ylim(-0.05, 1.05)
        betting_history = data["history"]
        sort = sorted(betting_history, key=lambda b: b["probability"])
        # if season == "2017":
        #     sort = list(filter(lambda b: b['probability'] <  0.8, sort))

        y = np.array([1 if b["bet"] == b["result"] else 0 for b in sort])
        x = [b["probability"] for b in sort]

        sp = csaps.UnivariateCubicSmoothingSpline(x, y, smooth=0.985)
        xs = np.linspace(x[0], x[-1], 150)
        ys = sp(xs)
        x_range = np.arange(min(x), max(x), 0.01)
        outcome_range = np.arange(0.1, 1, 0.01)
        ax.scatter(x, y, marker='x')
        ax.plot(outcome_range,outcome_range, 'r--')
        ax.plot(xs, ys, 'b-')
        ax.title.set_text(f"{season}/{int(season)+1}")
        ax.grid(True)
        subplotIdx += 1
    filename = f"{model_name.replace(' ', '')}-prediction-outcome-distribution"
    print(filename)
    plt.tight_layout()
    plt.savefig(f"plots/{filename}.png")
    plt.show()


def plot_prediction_success_all_preds(model_name, prob_outcome_dists):
    fig = plt.figure(figsize=(9, 4))
    subplotIdx = 1
    predictions = []
    for season, prob_outcomes in prob_outcome_dists.items():
        flattened_pred =[item for sublist in prob_outcomes for item in sublist]
        predictions += flattened_pred
    predictions = sorted(predictions, key=lambda p: p[0])
    x = [p[0] for p in predictions]
    y = [p[1] for p in predictions]
    print(f"MAX Y: {max(y)} | MIN Y: {min(y)}")
    sp = csaps.UnivariateCubicSmoothingSpline(x, y, smooth=0.9995)
    xs = np.linspace(x[0], x[-1], 150)
    ys = sp(xs)
    x_range = np.arange(min(x), 1, 0.01)
    plt.xlabel("Predicted probability")
    plt.ylabel("Outcome distribution")
    plt.scatter(x, y, marker='x', s=5)
    plt.plot(x_range,x_range, 'r--')
    plt.grid()
    plt.plot(xs, ys, 'b-')
    #plt.title(model_name)
    subplotIdx += 1
    filename = f"{model_name.replace(' ', '')}-prediction-outcome-distribution"
    print(filename)
    plt.tight_layout()
    plt.savefig(f"plots/{filename}.png")#, bbox_inches="tight")
    plt.show()

def plot_model_results(strategy_history, model_name, save_plot=True):
    for season, history in strategy_history.items():
        plt_season_results(history, model_name, season, save_plot=save_plot)


def plt_season_results(strategy_history, model_name, season, save_plot=True):
    fig = plt.figure(figsize=(11, 5))
    plot_idx = 1
    result_string = ""
    for name, history in strategy_history.items():
        bankrolls = [h[0] for h in history]
        bets = [h[1] for h in history]
        yields = list(map(bankroll_to_yield, bankrolls, bets))
        handles, labels, min_end, max_end, mean_end = draw_strategy_profit_graph(yields, len(yields[0]), name, plot_idx=plot_idx)
        result_string += name + f"\t\nMin: {min_end}" + f"\t\nMax: {max_end}" + f"\t\nMean: {mean_end}"
        plot_idx += 1

    f = open(f"plots/{model_name}-{season}-min-max-mean-YIELD.txt", "w")
    f.write(result_string)
    f.close()
    lgd = fig.axes[-1].legend(handles, labels, loc='lower center', bbox_to_anchor=(0.5, 0), ncol=3, fancybox=True, bbox_transform=fig.transFigure)
    plt.subplots_adjust(left=0.05, bottom=0.14, right=0.97, top=0.93, wspace=0.14, hspace=0.32)
    if save_plot:
        filename = f"{model_name.replace(' ', '')}-{season}"
        if not os.path.isdir(f"plots"):
            os.mkdir(f"plots")
        plt.savefig(f"plots/{filename}.png", bbox_extra_artists=(lgd,), bbox_inches="tight", dpi=fig.dpi)
    plt.show()


def draw_profit_graph(
        data, strategy_name, plot_idx=0, x_label='Game week', y_label='Return on investment', save_plots=False):
    #plt.figure(figsize=(11, 5))

    data_plot = plt.subplot(2, 2, plot_idx)

    data_plot.plot(data["x_axis"], data["Least Profitable"], c='red')
    data_plot.plot(data["x_axis"], data["Most Profitable"], c='green')
    data_plot.plot(data["x_axis"], data["Mean"], c='blue')

    #for i, (data_points, label) in enumerate(data):
    #    data_plot.plot(*data_points, label=label, ls=PLOT_STYLES[i][0], c=PLOT_STYLES[i][1])

    #plt.axis((0, 38, *plt.axis()[2:]))

    box = data_plot.get_position()

    data_plot.set_position([box.x0, box.y0, box.width * 0.775, box.height])

    # data_plot.set_xlabel(x_label)
    # data_plot.set_ylabel(y_label)

    #data_plot.set_xticks(np.arange(2, 38, 1), minor=True)

    data_plot.grid(which='minor', alpha=0.25)
    data_plot.grid(which='major', alpha=0.5)

    # data_plot.legend(
    #     loc='center left', bbox_to_anchor=(1.0, 0.5), fancybox=True, shadow=True, ncol=1
    # )

    data_plot.title.set_text(f"{strategy_name}")
    return data_plot.get_legend_handles_labels()


def draw_strategy_profit_graph(statistics, number_of_game_weeks, strategy_name, plot_idx = 0, save_plots=False):
    least_profitable, most_profitable, mean_return_on_investment = min_max_mean(statistics)
    print(f"{strategy_name} mean end: {mean_return_on_investment[-1]}")
    print(f"{strategy_name} max end: {most_profitable[-1]}")
    print(f"{strategy_name} min end: {least_profitable[-1]}")
    x_axis = np.arange(2, len(statistics[0]) + 2)

    data = {
        "x_axis": x_axis,
        "Least Profitable": least_profitable,
        "Most Profitable": most_profitable,
        "Mean": mean_return_on_investment
    }

    handles, labels = draw_profit_graph(data, strategy_name, plot_idx=plot_idx, save_plots=save_plots)
    return handles, labels, least_profitable[-1], most_profitable[-1], mean_return_on_investment[-1]

def min_max_mean(bankrolls, min_max_all=False):
    '''
    Return bankroll histories with min, max and mean end result
    :param bankrolls:
    :return: min, max and mean
    '''
    if min_max_all:
        min_br = [None] * len(bankrolls[0])
        max_br = [None] * len(bankrolls[0])
        for br in bankrolls:
            for i in range(len(br)):
                if min_br[i] is None or br[i] < min_br[i]:
                    min_br[i] = br[i]
                if max_br[i] is None or br[i] > max_br[i]:
                    max_br[i] = br[i]
        return min_br, max_br, np.mean(bankrolls, axis=0)

    sorted_statistics = sorted(bankrolls, key=lambda x: x[-1])

    return sorted_statistics[0], sorted_statistics[-1], np.mean(sorted_statistics, axis=0)


def bankroll_to_roi(bankroll):
    roi = []
    for i in range(len(bankroll)):
        roi.append((bankroll[i] - bankroll[0]) / bankroll[0])
    return roi


def bankroll_to_yield(bankroll, betting_history):
    roi = []
    for i in range(len(bankroll)):
        if sum([bet['bet_size'] for bet in betting_history[:i*10]]) is 0:
            roi.append(bankroll[i]-bankroll[0])
        else:
            roi.append((bankroll[i]-bankroll[0]) / (sum([bet['bet_size'] for bet in betting_history[:i*10]])))
    return roi
