import numpy as np

OUTCOMES = ['H', 'D', 'A']

class Strategy:
    def __init__(self, bankroll, min_confidence, min_probability):
        self.initial_bankroll = bankroll
        self.bankroll = bankroll
        self.min_confidence = min_confidence
        self.min_probability = min_probability
        self.bankroll_history = [bankroll]
        self.betting_history = []
        self.total_bets = 0
        self.games = 0

    def get_bet_size(self, odds, probability):
        raise NotImplementedError

    @property
    def name(self):
        raise NotImplementedError

    def get_bets(self, odds, probabilities):
        bets = [None] * 3
        best_bet_idx = None
        best_expected_profit = 0
        for i in range(3):
            expected_profit = odds[i] * probabilities[i]
            if expected_profit > self.min_confidence and probabilities[i] > self.min_probability and expected_profit > best_expected_profit:
                best_expected_profit = expected_profit
                best_bet_idx = i
                bets[i] = (odds[i], probabilities[i])
        return best_bet_idx, bets

    def evaluate_game_week(self, game_week, one_bet_only=False):
        self.games += 1
        profit = 0
        for idx in range(len(game_week['odds'])):
            bets = [None] * 3
            odds = game_week['odds'][idx]
            probabilities = game_week['probabilities'][idx].tolist()[0]
            if one_bet_only:
                best_bet_idx, bets = self.get_bets(odds, probabilities)
                if best_bet_idx is None:
                    continue
                self.total_bets += 1
                odd, probability = bets[best_bet_idx]

                bet_size = self.get_bet_size(odd, probability)

                self.bankroll -= bet_size
                if OUTCOMES[best_bet_idx] == game_week['results'][idx]:
                    profit += (bet_size * odd)
                self.add_bet_to_history(odd, probability, OUTCOMES[best_bet_idx], bet_size, game_week['results'][idx])
            else:
                _, bets = self.get_bets(odds, probabilities)
                for i, bet in enumerate(bets):
                    if bet is None:
                        continue
                    else:
                        self.total_bets += 1
                    odd, probability = bet
                    bet_size = self.get_bet_size(odd, probability)
                    self.bankroll -= bet_size
                    if OUTCOMES[i] == game_week['results'][idx]:
                        profit += (bet_size * odd)
                    print("Probabilities: " + str(probabilities))
                    # odds = get_best_odds(b_odds)
                    print("Odds: " + str(odds))
                    print("BET: " + OUTCOMES[i])
                    print("RESULT: " + game_week['results'][idx])
        self.bankroll += profit
        self.bankroll_history.append(self.bankroll)

    def add_bet_to_history(self, odd, prob, bet, bet_size, result):
        self.betting_history.append(
            {
                "odds": odd,
                "probability": prob,
                "bet": bet,
                "bet_size": bet_size,
                "result": result
             }
        )

    def evaluate_game(self, odds, probabilities, result):
        self.games += 1
        bets = [None] * 3
        profit = 0

        for i in range(3):
            expected_profit = odds[i] * probabilities[i]
            if expected_profit > self.min_confidence and probabilities[i] > self.min_probability:
                bets[i] = (odds[i], probabilities[i])

        for i, bet in enumerate(bets):
            if bet is None:
                continue
            else:
                self.total_bets += 1

            odd, probability = bet
            bet_size = self.get_bet_size(odd, probability)
            if OUTCOMES[i] == result:
                profit += (bet_size * odd) - bet_size
            else:
                profit -= bet_size
            print("BET: " + OUTCOMES[i])
            print("RESULT: " + result)
        self.bankroll += profit
        self.bankroll_history.append(self.bankroll)
        print("Bankroll for {} after game {}: {}".format(self.name, self.games, self.bankroll))

    def report(self):
        print("{name} placed {bets} bets on {games} games. End bankroll: {bankroll}".format(
            name=self.name,
            bets=self.total_bets,
            games=self.games,
            bankroll=self.bankroll
        ))

    def reset(self):
        self.bankroll = self.initial_bankroll
        self.bankroll_history = [self.bankroll]
        self.betting_history = []
        self.total_bets = 0
        self.games = 0


class FixedBet(Strategy):
    def __init__(self, bankroll, min_confidence, min_probability, bet_size):
        super().__init__(bankroll, min_confidence, min_probability)
        self.bet_size = bet_size

    def get_bet_size(self, odds, probability):
        return self.bet_size

    @property
    def name(self):
        return "FixedBet"

class FixedBetMostLikely(FixedBet):
    def evaluate_game_week(self, game_week, one_bet_only=True):
        self.games += 1
        profit = 0
        for idx, odd in enumerate(game_week['odds']):
            self.total_bets += 1
            bet_size = self.get_bet_size(odd, None)

            best_odd = min(odd)
            bet_idx = np.argmin(odd)
            self.add_bet_to_history(odd, None, OUTCOMES[bet_idx], bet_size, game_week['results'][idx])
            self.bankroll -= bet_size
            if OUTCOMES[bet_idx] == game_week['results'][idx]:
                profit += (bet_size * best_odd)
            print("BET: " + OUTCOMES[bet_idx])
            print("Odds: " + str(best_odd))
            print("RESULT: " + game_week['results'][idx])
        self.bankroll += profit
        self.bankroll_history.append(self.bankroll)
        print("Bankroll for {} after game week {}: {}".format(self.name, self.games, self.bankroll))

    @property
    def name(self):
        return "FixedBetMostLikely"


class FixedBetLeastLikely(FixedBet):
    def evaluate_game_week(self, game_week, one_bet_only=True):
        self.games += 1
        profit = 0
        for idx, odd in enumerate(game_week['odds']):
            self.total_bets += 1
            bet_size = self.get_bet_size(odd, None)
            best_odd = max(odd)
            bet_idx = np.argmax(odd)
            self.add_bet_to_history(odd, None, OUTCOMES[bet_idx], bet_size, game_week['results'][idx])
            self.bankroll -= bet_size
            if OUTCOMES[bet_idx] == game_week['results'][idx]:
                profit += (bet_size * best_odd)
            print("BET: " + OUTCOMES[bet_idx])
            print("Odds: " + str(best_odd))
            print("RESULT: " + game_week['results'][idx])
        self.bankroll += profit
        self.bankroll_history.append(self.bankroll)
        print("Bankroll for {} after game week {}: {}".format(self.name, self.games, self.bankroll))

    @property
    def name(self):
        return "FixedBetLeastLikely"


class FixedBetAlwaysHome(FixedBet):
    def evaluate_game_week(self, game_week, one_bet_only=True):
        self.games += 1
        profit = 0
        for idx, odd in enumerate(game_week['odds']):
            self.total_bets += 1
            bet_size = self.get_bet_size(odd, None)
            self.add_bet_to_history(odd, None, OUTCOMES[0], bet_size, game_week['results'][idx])
            self.bankroll -= bet_size
            if OUTCOMES[0] == game_week['results'][idx]:
                profit += (bet_size * odd[0])
            print("BET: " + OUTCOMES[0])
            print("Odds: " + str(odd[0]))
            print("RESULT: " + game_week['results'][idx])
        self.bankroll += profit
        self.bankroll_history.append(self.bankroll)
        print("Bankroll for {} after game week {}: {}".format(self.name, self.games, self.bankroll))

    @property
    def name(self):
        return "FixedBetAlwaysHome"


class FixedReturn(Strategy):
    def __init__(self, bankroll, min_confidence, min_probability, expected_return=1):
        super().__init__(bankroll, min_confidence, min_probability)
        self.expected_return = expected_return

    def get_bet_size(self, odds, probability):
        return self.expected_return / float(odds)

    @property
    def name(self):
        return "FixedReturn"


class KellyRatio(Strategy):
    def __init__(self, bankroll, min_confidence, min_probability, max_bet_factor=0.05):
        super().__init__(bankroll, min_confidence, min_probability)
        self.max_bet_factor = max_bet_factor

        self.max_bet_size = float(bankroll * max_bet_factor)

    @property
    def name(self):
        return 'KellyRatio'

    def get_bet_size(self, odds, probability):
        bet_size = self.bankroll * (probability * odds - 1) / (odds - 1)

        return min(bet_size, self.max_bet_size)


class VarianceAdjusted(Strategy):
    def __init__(self, bankroll, min_confidence, min_probability):
        super().__init__(bankroll, min_confidence, min_probability)

    @property
    def name(self):
        return 'VarianceAdjusted'

    def get_bet_size(self, odds, probability):
        return (2 * odds * (1 - probability)) ** (-1)

