import os
import pickle
import re

import keras
import numpy as np
import pandas
import tensorflow as tf

import models.model_evaluator as model_evaluator
import util
from betting.strategy import FixedBet, FixedReturn, KellyRatio, VarianceAdjusted
from datasets.dataloading import separate_training_validation_data_spi, \
    separate_training_validation_data, separate_test_data, separate_training_validation_data_exp, load_dataframe_data
from models.expected_lineup import ExpectedLineUp
from models.soccerpowerindexmodel import SoccerPowerIndex
from models.sub_expected_lineup import SubExpectedLineup
from models.sub_soccerpowerindexmodel import SubSoccerPowerIndex
from util.model_input_util import create_multiple_inputs_targets_sub_net_dataframe, \
    create_features_multiple_input_dataframe, \
    create_multiple_target_output_dataframe, create_features_multiple_input, create_multiple_target_output, \
    create_multiple_inputs_targets_sub_net_peewee

config = tf.ConfigProto( device_count = {'GPU': 1 , 'CPU': 8} )
sess = tf.Session(config=config)
keras.backend.set_session(sess)

EXP_FOLDER = "outputs"
EXP_NAME = "experiment"

layer_sizes = [8, 16, 32, 64]
layers = [1, 2, 3]
size_configs = [ [x] * n_layers for x in layer_sizes for n_layers in layers]
activation_functions = ['relu', 'sigmoid', 'tanh']
models = [SoccerPowerIndex, SubSoccerPowerIndex, ExpectedLineUp, SubExpectedLineup]
NUM_EPOCHS = 500
validation_split = 0.1
INIT_BANKROLL = 100
MIN_CONFIDENCE = 1.1
MIN_PROBABILITY = 0.1
NUM_MODEL_RUNS = 5
one_bet_only = True

strategies = [
    FixedBet(INIT_BANKROLL, MIN_CONFIDENCE, MIN_PROBABILITY, bet_size=1),
    FixedReturn(INIT_BANKROLL, MIN_CONFIDENCE, MIN_PROBABILITY, expected_return=1),
    KellyRatio(INIT_BANKROLL, MIN_CONFIDENCE, MIN_PROBABILITY),
    VarianceAdjusted(INIT_BANKROLL, MIN_CONFIDENCE, MIN_PROBABILITY)
]


def getPiData(df_data):
    df_data = df_data.sample(frac=1)
    fatigue_input, psychology_input, strength_input, form_input, targets = create_multiple_inputs_targets_sub_net_dataframe(df_data,
                                                                                                                            SubSoccerPowerIndex.get_default_features(),
                                                                                                                            SubSoccerPowerIndex.get_default_features()['feature_list'])

    all_data = [(fatigue_input[i], psychology_input[i], strength_input[i], form_input[i], targets[i])
                for i in
                range(len(fatigue_input))]

    validation_data, training_data, validation_targets, training_targets = separate_training_validation_data_spi(
        all_data, fatigue_input, form_input, psychology_input, strength_input, targets, validation_split)

    return [validation_data, validation_targets, training_data, training_targets]


def getSpiData(data, feature_list):
    inputs = create_features_multiple_input_dataframe(data, feature_list)
    targets = create_multiple_target_output_dataframe(data, feature_list)

    if len(inputs) != len(targets):
        print("Inconsistent input and target vectors. Len not matching. Exiting.")
        raise Exception

    all_data = [(inputs[i], targets[i]) for i in range(len(inputs))]
    return [separate_training_validation_data(all_data, validation_split)]


def getExpectedLineupData(data, feature_dict):
    inputs = create_features_multiple_input(data, feature_dict)
    targets = create_multiple_target_output(data)
    if len(inputs) != len(targets):
        print("Inconsistent input and target vectors. Len not matching. Exiting.")
        raise Exception

    all_data = [(inputs[i], targets[i]) for i in range(len(inputs))]
    np.random.shuffle(all_data)
    return [separate_training_validation_data(all_data, validation_split)]


def getSubExpectedLineupData(data, feature_dict):
    fatigue_input, strength_input, form_input, targets = create_multiple_inputs_targets_sub_net_peewee(data, feature_dict)

    if len(fatigue_input) != len(targets):
        print("Inconsistent input and target vectors. Len not matching. Exiting.")
        raise Exception

    all_data = [(fatigue_input[i], strength_input[i], form_input[i], targets[i])
                for i in
                range(len(fatigue_input))]
    np.random.shuffle(all_data)
    validation_data, training_data, validation_targets, training_targets = separate_training_validation_data_exp(
        all_data, fatigue_input, strength_input, form_input, targets, validation_split)
    return [validation_data, validation_targets, training_data, training_targets]


def load_model_filenames():
    filenames = []
    for dirName, subdirList, fileList in os.walk(EXP_FOLDER):
        print('Found directory: %s' % dirName)
        for filename in fileList:
            if re.search('^.*FINAL.*$', filename):
                filenames.append(dirName+ "/" + filename)
    print("Loading filenames for  %s models" % str(len(filenames)))
    return filenames



if __name__ == "__main__":
    test_league_seasons = {
        "leagues": ["EPL"],
        "seasons": ["2016", "2017"]
    }


    peewee_all_data = pickle.load(open("../models/data_files/all-data-peewee-n_5-player-list", "rb"))
    peewee_training_data, peewee_test_data = separate_test_data(peewee_all_data, test_league_seasons)
    df_data, df_test_data = load_dataframe_data(test_league_seasons, "../models/data_files/all-data-df")
    # model_data = {
    #      "PiFootball": getPiData(df_data),
    #      "SoccerPowerIndex": getSpiData(df_data, SoccerPowerIndex.get_default_features()['feature_list']),
    #      "ExpectedLineup": getExpectedLineupData(peewee_training_data, ExpectedLineUp.get_default_features())
    # }
    # sub_exp_model_data = {
    #      "SubExpectedLineup": getSubExpectedLineupData(peewee_training_data, SubExpectedLineup.get_default_features())
    # }
    #print("Storing dict")
    #pickle.dump(df_test_data, open("df_test_data.df", "wb"))
    #pickle.dump(peewee_test_data, open("pw_test_data.pw", "wb"))
    #util.util.save_dict_to_file("sub_exp_model_data_dict.npy", sub_exp_model_data)
    #print("Dict stored")
    model_data = util.util.load_dict_from_file("model_data_dict.npy")
    sub_exp_model_data = util.util.load_dict_from_file("sub_exp_model_data_dict.npy")
    #df_test_data = pickle.load(open("df_test_data.df", "rb"))
    #pw_test_data = pickle.load(open("pw_test_data.pw", "rb"))

    test_log_df = pandas.DataFrame(columns=[
        "model",
        "index_run",
        "size_config",
        "activation",
        "max_val_acc",
        "min_val_loss",
        "val_accuracy",
        "val_rps_min",
        "val_rps_max",
        "val_rps_mean",
        "fixed_bet_roi-2016",
        "fixed_return_roi-2016",
        "kelly_roi-2016",
        "variance_roi-2016",
        "fixed_bet_roi-2017",
        "fixed_return_roi-2017",
        "kelly_roi-2017",
        "variance_roi-2017"
    ])

    for m in models:
        for size_config in size_configs:
            for act_func in activation_functions:
                for i in range(NUM_MODEL_RUNS):
                    print("Model_{}_act_{}_size_config{}_NUMBER_{}".format(str(m),act_func, str(size_config), str(i)))
                    if m == SubSoccerPowerIndex:
                        model = SubSoccerPowerIndex(size_config,
                                                    [len(SubSoccerPowerIndex.get_default_features()['fatigue_factors']),
                                                    len(SubSoccerPowerIndex.get_default_features()['form_factors']),
                                                    len(SubSoccerPowerIndex.get_default_features()['strength_factors']),
                                                    len(SubSoccerPowerIndex.get_default_features()['psychology_factors'])],
                                                    size_config, model_number=i, activation_function=act_func)
                        history = model.model.fit(x=model_data['PiFootball'][2],
                                                      y=model_data['PiFootball'][3],
                                                      epochs=NUM_EPOCHS,
                                                      validation_data=(model_data['PiFootball'][0], model_data['PiFootball'][1]),
                                                  shuffle=True,
                                                  verbose=2,
                                                  callbacks=[model.checkpoint])

                        min_rps, max_rps, mean_rps, val_acc = model_evaluator.evaluate_validation_pi_data([model_data['PiFootball'][0], model_data['PiFootball'][1]], model.model)
                        #strategy_bankrolls = evaluate_games_weekly(model.model, test_league_seasons, df_test_data, m.get_default_features(), strategies, dataframe=True, one_bet_only=one_bet_only)

                    elif m == ExpectedLineUp:
                        model = ExpectedLineUp(size_config, 50,
                                               model_number=i, activation_function=act_func)
                        history = model.train(model_data['ExpectedLineup'][0][0], model_data['ExpectedLineup'][0][1], epochs=NUM_EPOCHS)
                        min_rps, max_rps, mean_rps, val_acc = model_evaluator.evaluate_validation_exp_data(model_data['ExpectedLineup'][0][1], model.model)
                        #strategy_bankrolls = evaluate_games_weekly(model.model, test_league_seasons, pw_test_data, m.get_default_features(), strategies, dataframe=False, one_bet_only=one_bet_only)
                    elif m == SoccerPowerIndex:
                        model = SoccerPowerIndex(size_config,
                                                 len(SoccerPowerIndex.get_default_features()['feature_list']), model_number=i,
                                                 activation_function=act_func)
                        history = model.train(model_data['SoccerPowerIndex'][0][0], model_data['SoccerPowerIndex'][0][1], epochs=NUM_EPOCHS)
                        min_rps, max_rps, mean_rps, val_acc = model_evaluator.evaluate_validation_exp_data(model_data['SoccerPowerIndex'][0][1], model.model)
                        #strategy_bankrolls = evaluate_games_weekly(model.model, test_league_seasons, df_test_data, m.get_default_features(), strategies, dataframe=True,one_bet_only=one_bet_only)
                    elif m == SubExpectedLineup:
                        model = SubExpectedLineup(size_config,
                                                 [2,24,24], size_config, model_number=i,
                                                 activation_function=act_func)
                        try:
                            history = model.model.fit(x=sub_exp_model_data['SubExpectedLineup'][2],
                                              y=sub_exp_model_data['SubExpectedLineup'][3],
                                              epochs=NUM_EPOCHS,
                                              validation_data=(sub_exp_model_data['SubExpectedLineup'][0], sub_exp_model_data['SubExpectedLineup'][1]),
                                              shuffle=True,
                                              verbose=2,
                                              callbacks=[model.checkpoint])
                        except KeyboardInterrupt:
                            min_rps, max_rps, mean_rps, val_acc = model_evaluator.evaluate_validation_sub_exp_data([sub_exp_model_data['SubExpectedLineup'][0], sub_exp_model_data['SubExpectedLineup'][1]], model.model)
                        #strategy_bankrolls = evaluate_games_weekly(model.model, test_league_seasons, df_test_data, m.get_default_features(), strategies, dataframe=True,one_bet_only=one_bet_only)

                    min_val_loss_idx = np.argmin(history.history['val_loss'])
                    test_log_df = test_log_df.append({
                        "model": model.model_name,
                        "index_run": model.model_number,
                        "size_config": size_config,
                        "activation": act_func,
                        "max_val_acc": history.history['val_acc'][min_val_loss_idx],
                        "min_val_loss": history.history['val_loss'][min_val_loss_idx],
                        "val_accuracy": val_acc,
                        "val_rps_min": min_rps,
                        "val_rps_max": max_rps,
                        "val_rps_mean": mean_rps
                    }, ignore_index=True)
                    model.save_model()

    test_log_df.to_csv(f"{EXP_NAME}_{NUM_EPOCHS}_epochs.csv")
