from tensorflow.python.keras import Sequential, optimizers
from tensorflow.python.keras.layers import Dense, Dropout
from tensorflow.python.keras.models import load_model
from tensorflow.python.keras.callbacks import TensorBoard, ModelCheckpoint
from time import time
import numpy as np

from util.util import rps

OPTIMIZERS = {
    "adam": optimizers.Adam,
    "adagrad": optimizers.Adagrad,
    "sgd": optimizers.SGD,
    "rmsprop": optimizers.RMSprop
}


class KerasModel:

    def __init__(self, network_dims, input_size, dir_name, model_name, model_number, batch_size=256, activation_function='relu', optimizer='adam', learning_rate=0.0001, loss="categorical_crossentropy"):
        self.network_dims = network_dims
        self.activation_function = activation_function
        self.optimizer = optimizer
        self.lr = learning_rate
        self.loss = loss
        self.input_size = input_size
        self.batch_size = batch_size
        self.model_number = model_number

        self.dir_name = dir_name

        self.model = Sequential()
        self.model.add(Dense(network_dims[0], input_shape=(input_size,), activation=activation_function))
        for dim in network_dims[1:]:
            self.model.add(Dense(dim, activation=activation_function))
            self.model.add(Dropout(0.05))
        self.model.add(Dense(3, activation='softmax'))
        self.model.compile(optimizer=OPTIMIZERS[optimizer](lr=learning_rate),
                           loss=loss,
                           metrics=['accuracy', 'top_k_categorical_accuracy'])

        timestamp = time()
        self.tensorboard = TensorBoard(log_dir="./outputs/{}/{}-logs".format(dir_name, self.parameter_string), batch_size=batch_size)
        self.checkpoint = ModelCheckpoint("./outputs/{}/{}-model".format(dir_name, self.parameter_string), save_best_only=True, monitor="val_loss")

    @property
    def parameter_string(self):
        return "dims={dims}_opt={opt}_act={act}_lr={lr}_loss={loss}_run{model_number}".format(
            dims="-".join(str(d) for d in self.network_dims),
            opt=self.optimizer,
            act=self.activation_function,
            lr=self.lr,
            loss=self.loss,
            model_number=self.model_number
        )

    def load_model(self, filename):
        self.model = load_model(filename)

    def forward(self, input):
        return self.model.predict(input)

    def evaluate(self, test_data, test_labels):
        return self.model.evaluate(x=test_data, y=test_labels)

    def train(self, training_data, validation_data, epochs=1000):
        inputs = np.array([data[0] for data in training_data])
        targets = np.array([data[1] for data in training_data])
        validation_x = np.array([data[0] for data in validation_data])
        validation_y = np.array([data[1] for data in validation_data])
        return self.model.fit(x=inputs, y=targets,
                              batch_size=self.batch_size,
                              epochs=epochs,
                              validation_data=(validation_x, validation_y),
                              shuffle=True,
                              verbose=2,
                              callbacks=[self.checkpoint])

    def save_model(self):
        self.model.save("./outputs/{}/{}-FINAL-model".format(self.dir_name, self.parameter_string))