from time import time

from keras import Input, Model, optimizers
from keras.callbacks import ModelCheckpoint
from keras.layers import Dense, Dropout, concatenate

OPTIMIZERS = {
    "adam": optimizers.Adam,
    "adagrad": optimizers.Adagrad,
    "sgd": optimizers.SGD,
    "rmsprop": optimizers.RMSprop
}

class SubNetworkModel():
    def __init__(self, sub_input_network_dims, factor_input_sizes,
                 super_network_dims, model_number,dir_name="sub-pi-football", model_name="sub-pi-football", loss="categorical_crossentropy",
                 activation_function="relu", batch_size=256, optimizer='adam', learning_rate=0.0001):

        self.sub_dims = sub_input_network_dims
        self.super_dims = super_network_dims
        self.activation_function = activation_function
        self.optimizer = optimizer
        self.loss = loss
        self.lr = learning_rate
        self.batch_size = batch_size
        self.model_number = model_number
        branch_inputs = []
        branch_outputs = []
        for factor in factor_input_sizes:
            factor_branch_input = Input(shape=(factor,))
            branch_inputs.append(factor_branch_input)
            factor_branch_output = Dense(super_network_dims[0], activation=activation_function)(factor_branch_input)
            factor_branch_output = Dropout(0.05)(factor_branch_output)
            for dim in sub_input_network_dims[1:]:
                factor_branch_output = Dense(dim, activation=activation_function)(factor_branch_output)
                factor_branch_output = Dropout(0.05)(factor_branch_output)
            branch_outputs.append(factor_branch_output)
        merge = concatenate(
            branch_outputs)
        for dim in super_network_dims:
            merge = Dense(dim, activation=activation_function)(merge)
            merge = Dropout(0.05)(merge)
        output = Dense(3, activation='softmax')(merge)
        self.model = Model(
            inputs=branch_inputs,
            outputs=[output])
        print(self.model.summary())

        self.model.compile(optimizer=OPTIMIZERS[optimizer](lr=learning_rate),
                           loss=loss,
                           metrics=['accuracy'])
        timestamp = time()
        self.model_name = model_name + str(timestamp)
        self.dir_name = dir_name
        #self.tensorboard = TensorBoard(log_dir="./outputs/{}/{}-logs".format(dir_name, self.parameter_string),
        #                               batch_size=batch_size)
        self.checkpoint = ModelCheckpoint("./outputs/{}/{}-model".format(dir_name, self.parameter_string),
                                          save_best_only=True, monitor="val_loss")


    @property
    def parameter_string(self):
        return "dims={dims}_opt={opt}_act={act}_lr={lr}_loss={loss}_run{model_number}".format(
            dims="-".join(str(d) for d in self.sub_dims),
            opt=self.optimizer,
            act=self.activation_function,
            lr=self.lr,
            loss=self.loss,
            model_number=self.model_number
        )


    def save_model(self):
        self.model.save("./outputs/{}/{}-FINAL-model".format(self.dir_name, self.parameter_string))