import math

import numpy as np
from gym.spaces import Box, Discrete
from peewee import fn, JOIN

import database.dbmodels as db

MAX_EPISODE_STEPS = 400


class WeeklyEnvironment:
    action_space = Discrete(4)
    observation_space = Box(
        low=np.array([
            0.0, # Home prob
            0.0, # Draw prob
            0.0, # Away prob
            0.0, # Home odds
            0.0, # Draw odds
            0.0, # Away odds
            0, # Matches before reward
            -float("inf"), # Expected profit of pending bets
            -float("inf") # Money
        ]),
        high=np.array([
            1.0, # Home prob
            1.0, # Draw prob
            1.0, # Away prob
            100, # Home odds
            100, # Draw odds
            100, # Away Odds
            10, # Matches before reward
            float("inf"), # Expected profit of pending bets
            float("inf") # Money,
        ]))

    MONEY_SCALE_FACTOR = 100

    def __init__(self, initial_bankroll, bet_size, fake_data=True, matches_between_rewards=1):
        self.matches_original = list(db.Match.select()
            .where((db.Match.spi_prob_a.is_null(False)))
            .join(db.BookmakerMatchOdds, JOIN.LEFT_OUTER)
            .group_by(db.Match)
            .having(fn.COUNT(db.BookmakerMatchOdds.bookmaker) > 0))
        print(f"Loaded {len(self.matches_original)} matches from DB")
        self.load_matches()

        self.fake_data = fake_data
        self.initial_money = initial_bankroll
        self.bet_size = bet_size
        self.matches_between_rewards = matches_between_rewards
        self.matches_before_reward = matches_between_rewards
        self.pending_rewards = []
        self.reset()

    def load_matches(self):
        self.matches = []
        for match in self.matches_original:
            wh_odds = list(filter(lambda b: b.bookmaker.name == "William Hill", match.odds))
            if len(wh_odds) == 0:
                print("No William Hill odds")
                continue
            wh_odds = wh_odds[0]
            if match.h_goals > match.a_goals:
                outcome = 'h'
            elif match.h_goals < match.a_goals:
                outcome = 'a'
            else:
                outcome = 'd'
            self.matches.append({
                "probs": (match.spi_prob_h, match.spi_prob_tie, match.spi_prob_a),
                "odds": (wh_odds.home, wh_odds.draw, wh_odds.away),
                "outcome": outcome
            })

    def get_match_outcome(self):
        if self.fake_data:
            probs = self.state[0:3]
            return np.random.choice(["h", "d", "a"], p=probs)
        return self.match["outcome"]

    def get_action_outcome_info(self, action):
        probs = self.state[:3]
        odds = self.state[3:-1]

        if action == 0:
            # No bet
            return "n", 1, 1
        elif action == 1:
            return "h", odds[0], probs[0]
        elif action == 2:
            return "d", odds[1], probs[1]
        else:
            return "a", odds[2], probs[2]


    def log_profit_reward(self, profit):
        return 0 if profit == 0 else math.copysign(1, profit) * math.log10(abs(profit))

    def log_bankroll_reward(self, money):
        return -10 if money <= 0 else math.log10(money)

    def log_bankroll_end_reward(self, money, done):
        return 0 if not done or money == 0 else math.copysign(1, money) * math.log10(abs(money))

    def exp_gain_reward(self, action):
        if action == 0:
            return 0
        expected_profit = self.state[action - 1] * self.state[(action - 1) + 3]
        return expected_profit - 1

    def feasibility_reward(self, action):
        if action == 0:
            return 0
        expected_profit = self.state[action - 1] * self.state[(action - 1) + 3]
        if expected_profit > 1:
            return 100
        else:
            return -100

    def best_feasible_bet(self):
        best_bet = None
        outcomes = ["h", "d", "a"]
        for i in range(3):
            exp_profit = self.state[i] * self.state[i+3]
            if exp_profit < 1:
                continue
            if best_bet is None or exp_profit > best_bet[1] * best_bet[2]:
                best_bet = (outcomes[i], self.state[i], self.state[i+3])
        if best_bet is None:
            return ("n", 0, 0)
        return best_bet

    def step(self, action):
        outcome = self.get_match_outcome()
        bet_size_modifier = 1
        bet_outcome, bet_odds, bet_probability = self.get_action_outcome_info(action)

        money = self.state[-1]
        bet_size = self.bet_size * bet_size_modifier
        exp_profit = self.state[-2]

        if bet_outcome == 'n':
            profit = 0
        elif bet_outcome == outcome:
            profit = bet_size * bet_odds
            money -= bet_size
            self.pending_rewards.append(profit)
            exp_profit += bet_size * bet_probability
        else:
            self.pending_rewards.append(-bet_size)
            exp_profit += bet_size * bet_probability
            profit = -bet_size
            money -= bet_size

        best_bet = self.best_feasible_bet()
        if best_bet[0] == outcome:
            optimal_bet_reward = best_bet[-1]
        elif best_bet[0] == 'n':
            optimal_bet_reward = 0
        else:
            optimal_bet_reward = -1
        self.optimal_rewards.append(optimal_bet_reward)

        self.n_match += 1
        self.matches_before_reward -= 1
        if self.matches_before_reward == 0:
            money += sum(filter(lambda r: r > 0, self.pending_rewards))
            reward = sum(self.pending_rewards)
            optimal_reward = sum(self.optimal_rewards)
            self.matches_before_reward = self.matches_between_rewards
            self.pending_rewards = []
            self.optimal_rewards = []
            exp_profit = 0
        elif money <= 0:
            reward = -10
            optimal_reward = sum(self.optimal_rewards)
        else:
            reward = 0
            optimal_reward = 0
        done = money <= 0
        self.state = self.get_state_vector(money, exp_profit)

        return self.state, reward, done, {"profit": profit, "optimal": optimal_reward}

    def reset(self):
        self.match = np.random.choice(self.matches)
        self.n_match = 0
        self.state = self.get_state_vector(self.initial_money, 0)
        self.last_action = None
        self.pending_rewards = []
        self.optimal_rewards = []
        return np.array(self.state)


    def get_state_vector(self, money, exp_profit):
        if self.fake_data:
            probs = np.random.uniform(0, 1, 3)
            probs = probs / sum(probs)
            odds = np.random.uniform(0.1, 0.7, 3)
            odds = 1 / (odds / sum(odds))
            return (
                probs[0], probs[1], probs[2],
                odds[0], odds[1], odds[2],
                self.matches_before_reward, exp_profit, money
            )
        self.match = self.matches[np.random.randint(len(self.matches))]
        return self.match["probs"] + self.match["odds"] + (self.matches_before_reward, exp_profit, money)

class SingleMatchEnvironment:
    # No bet, H, D, A
    action_space = Discrete(4)
    observation_space = Box(
        low=np.array([
            0.0, # Home prob
            0.0, # Draw prob
            0.0, # Away prob
            0.0, # Home odds
            0.0, # Draw odds
            0.0, # Away odds
            -float("inf") # Money
        ]),
        high=np.array([
            1.0, # Home prob
            1.0, # Draw prob
            1.0, # Away prob
            100, # Home odds
            100, # Draw odds
            100, # Away Odds
            float("inf") # Money,
        ]))

    MONEY_SCALE_FACTOR = 100

    def __init__(self, initial_bankroll, bet_size, fake_data=True):
        self.matches_original = list(db.Match.select()
                                     .where((db.Match.spi_prob_a.is_null(False)))
                                     .join(db.BookmakerMatchOdds, JOIN.LEFT_OUTER)
                                     .group_by(db.Match)
                                     .having(fn.COUNT(db.BookmakerMatchOdds.bookmaker) > 0))
        print(f"Loaded {len(self.matches_original)} matches from DB")
        self.load_matches()

        self.fake_data = fake_data
        self.initial_money = initial_bankroll
        self.bet_size = bet_size
        self.reset()

    def load_matches(self):
        self.matches = []
        for match in self.matches_original:
            wh_odds = list(filter(lambda b: b.bookmaker.name == "William Hill", match.odds))
            if len(wh_odds) == 0:
                print("No William Hill odds")
                continue
            wh_odds = wh_odds[0]
            if match.h_goals > match.a_goals:
                outcome = 'h'
            elif match.h_goals < match.a_goals:
                outcome = 'a'
            else:
                outcome = 'd'
            self.matches.append({
                "probs": (match.spi_prob_h, match.spi_prob_tie, match.spi_prob_a),
                "odds": (wh_odds.home, wh_odds.draw, wh_odds.away),
                "outcome": outcome
            })

    def get_match_outcome(self):
        if self.fake_data:
            probs = self.state[0:3]
            return np.random.choice(["h", "d", "a"], p=probs)
        return self.match["outcome"]

    def get_action_outcome_info(self, action):
        probs = self.state[:3]
        odds = self.state[3:-1]

        if action == 0:
            # No bet
            return "n", 1, 1
        elif action == 1:
            return "h", odds[0], probs[0]
        elif action == 2:
            return "d", odds[1], probs[1]
        else:
            return "a", odds[2], probs[2]

    def log_profit_reward(self, profit):
        return 0 if profit == 0 else math.copysign(1, profit) * math.log10(abs(profit))

    def log_bankroll_reward(self, money):
        return -10 if money <= 0 else math.log10(money)

    def log_bankroll_end_reward(self, money, done):
        return 0 if not done or money == 0 else math.copysign(1, money) * math.log10(abs(money))

    def exp_gain_reward(self, action):
        if action == 0:
            return 0
        expected_profit = self.state[action - 1] * self.state[(action - 1) + 3]
        return expected_profit - 1

    def feasibility_reward(self, action):
        if action == 0:
            return 0
        expected_profit = self.state[action - 1] * self.state[(action - 1) + 3]
        if expected_profit > 1:
            return 100
        else:
            return -100

    def best_feasible_bet(self):
        best_bet = None
        outcomes = ["h", "d", "a"]
        for i in range(3):
            exp_profit = self.state[i] * self.state[i+3]
            if exp_profit < 1:
                continue
            if best_bet is None or exp_profit > best_bet[1] * best_bet[2]:
                best_bet = (outcomes[i], self.state[i], self.state[i+3])
        if best_bet is None:
            return ("n", 0, 0)
        return best_bet

    def step(self, action):
        outcome = self.get_match_outcome()
        bet_size_modifier = 1
        bet_outcome, bet_odds, bet_probability = self.get_action_outcome_info(action)

        money = self.state[-1]
        bet_size = self.bet_size * bet_size_modifier

        if bet_outcome == 'n':
            profit = 0
        elif bet_outcome == outcome:
            profit = bet_size * bet_odds
            money -= bet_size
        else:
            profit = -bet_size
            money -= bet_size

        best_bet = self.best_feasible_bet()
        if best_bet[0] == outcome:
            optimal_reward = best_bet[-1]
        elif best_bet[0] == 'n':
            optimal_reward = 0
        else:
            optimal_reward = -1

        self.n_match += 1
        money += profit
        #optimal_reward = self.best_feasible_bet()[-1]
        if money > 0:
            reward = profit
        elif money <= 0:
            reward = -10
        done = money <= 0
        self.state = self.get_state_vector(money)

        return self.state, reward, done, {"profit": profit, "optimal": optimal_reward}

    def reset(self):
        self.match = np.random.choice(self.matches)
        self.n_match = 0
        self.state = self.get_state_vector(self.initial_money)
        self.last_action = None
        self.pending_rewards = []
        self.optimal_rewards = []
        return np.array(self.state)

    def get_state_vector(self, money):
        if self.fake_data:
            probs = np.random.uniform(0, 1, 3)
            probs = probs / sum(probs)
            odds = np.random.uniform(0.1, 0.7, 3)
            odds = 1 / (odds / sum(odds))
            return (
                probs[0], probs[1], probs[2],
                odds[0], odds[1], odds[2],
                money
            )
        self.match = self.matches[np.random.randint(len(self.matches))]
        return self.match["probs"] + self.match["odds"] + (money, )
